/////////////////////////////////////////////////
//Revolution Engine.
//by Technik
/////////////////////////////////////////////////

#ifndef REV_H
#define REV_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <ogcsys.h>
#include <gccore.h>
#include <math.h>
#include <wiiuse/wpad.h>//inclusiones de libogc
#include <fat.h>//Para poder trabajar con el sistema de archivos
#include <pngu.h>//Para las texturas en formato png
#include <time.h>
#include <ogc/lwp_watchdog.h>

typedef struct REV_Object{//Object, objeto en el espacio.
struct REV_Mesh * Mesh;//Modelo 3D En caso de ser estatico
struct _MD5modelo * Md5;//Modelo 3D En caso de ser animado
struct REV_Object * Next;//Puntero al siguiente objeto de la lista.
struct REV_Ph * State;//Sus datos fisicos concretos.
struct REV_Material * Material;//Material del modelo, con informacion de colores, alpha, etc...
struct REV_Body * Body;//Informacion fisica. Compartida.
GXColor baseColor;
u8 Visible;
} OBJECT;

typedef struct REV_Light{
GXLightObj * lit;
GXColor color;
Vector Pos;
} LIGHT;

typedef struct REV_Root{//Root
struct REV_Object * PhnObj;//Lista de objetos fantasma
struct REV_Object * StcObj;//Lista de objetos estaticos
struct REV_Object * DyUObj;//Lista de objetos dinamicos inactivos
struct REV_Object * DyAObj;//Lista de objetos dinamicos activos
struct REV_Object * ManObj;//Lista de objetos animados manualmente
struct REV_Object * ElsObj;//Lista con todos los objects que carecen de Body.
void * FstPnl;
struct REV_Light * lights;
} ROOT;

#include "REVSpace.h"//Para tratamiento de datos en el espacio.
#include "REVMd5.h"
#include "REVMesh.h"//Para el tratamiento de 3D
#include "REVdefs.h"//definiciones que pueden resultar utiles
#include "REVCams.h"//Para usar las funciones de camara
#include "REVPhysics.h"
#include "REVMaterial.h"
#include "REVImage.h"
#include "REVScreen.h"
#include "REVFreetype.h"
#include "REV3Dprimitives.h"

#define DEFAULT_FIFO_SIZE	(256*1024)


//Funciones

/****** NewRoot ******
//Descripcion: Crea una nueva raiz que el motor puede utilizar, de momento no se soportan multiples raices
//Argumentos: Ninguno
//Devuelve: Puntero a la nueva raiz creada*/
struct REV_Light * newLight(Vector Pos, GXColor color);

struct REV_Root * newRoot();

void DelRoot(ROOT * root);

void addObj(OBJECT * object);

struct REV_Object * NewObj(struct REV_Mesh * mesh, struct REV_Material * material, struct REV_Body * Body,
															Vector Pos, Vector Ang, u8 Visible);

u8 RemoveObjFromList(OBJECT * List, OBJECT * object);//Quita un objeto de la lista dada.
void AddObjToList(OBJECT * List, OBJECT * object);//Aade un objeto a la lista.

void FreeObj(OBJECT * object);

void clearRoot(ROOT * root);
void REV_Init();
void draw3D_init();
void draw2D_init();
void REV_Process(CAMERA * mCamera);

#endif
