/////////////////////////////////////////////////
//Revolution Engine.
//by Technik
//Md5 Loading and animation code by KaWeNGoD
/////////////////////////////////////////////////

#ifndef REVMD5_H
#define REVMD5_H

#include "REV.h"

#define MD5_VERSION		10

// Vectores
typedef float MD5vec2[2];
typedef float MD5vec3[3];

// Quaterniones (x, y, z, w)
typedef float MD5quat4[4];

enum {
  X, Y, Z, W
};

// Union
typedef struct _MD5union
{
	char nombre[64];	// Nombre del joint
	int pariente;		// Indice de union

	MD5vec3 pos;		// Posicion de la union
	MD5quat4 orient;	// Quaternion de orientacion de la union

} MD5union;

// Vertice
typedef struct _MD5vertice
{
	MD5vec2 st;		// Coordenadas de textura del vertice

	int inicio;		// Peso inicial para calcular la posicion final
	int total;		// Cantidad de pesos para calcular la posicion final

} MD5vertice;

// Triangulo
typedef struct _MD5triangulo
{
	int indice[3];		// Indice de los vertices del triangulo

} MD5triangulo;

// Peso
typedef struct _MD5peso
{
	int uniones;
	float bias;

	MD5vec3 pos;

} MD5peso;

// Bounding box
typedef struct _MD5caja
{
	MD5vec3 min;
	MD5vec3 max;

} MD5caja;

// MD5 mesh
typedef struct _MD5mesh
{
	MD5vertice 	*vertices;
	MD5triangulo 	*triangulos;
	MD5peso 	*pesos;
	MD5vec3		*norm_inic;

	f32	*Verts;			// Array de vertices
	f32	*Texcoords;		// Array de coordenadas de textura
	f32	*Normales;		// Array de normales

	u16 *VIndices;			// Indices de vertices

	int num_vertices;
	int num_triangulos;
	int num_pesos;

	char 		shader [256];	// Nombre de la textura del modelo ( usado para cargar la textura )

} MD5mesh;

// Datos de animacion
typedef struct _MD5anim
{
	int num_frames;
	int num_uniones;
	int frameRate;

	MD5union **EsqueletoFrames;
	MD5caja *cajas;

} MD5anim;

// Informacion de la animacion
typedef struct _MD5infoanim
{
	int curr_frame;
	int next_frame;

	float last_time;
	float max_time;

} MD5infoanim;

// Informacion de una union
typedef struct _MD5infounion
{
	char nombre[64];
	int pariente;
	int flags;
	int IndiceInicio;

} MD5infounion;

/* Union de frame base */
typedef struct _MD5unionbase
{
	MD5vec3 pos;
	MD5quat4 orient;

} MD5unionbase;

// Estructura de un modelo MD5
typedef struct _MD5modelo
{
	MD5union *baseEsqueleto;
	MD5mesh *meshes;
	MD5anim *animacion;	// Una animacion
	MD5infoanim *Infoanim;	// Informacion de las animaciones

	int num_uniones;
	int num_meshes;
	int num_animaciones;

} MD5modelo;

/*
 * Funciones de los quaterniones
 */
void Quat_calculoW ( MD5quat4 q );
void Quat_normalizar ( MD5quat4 q );
void Quat_multQuat ( const MD5quat4 qa, const MD5quat4 qb, MD5quat4 salida );
void Quat_multVec ( const MD5quat4 q, const MD5vec3 v, MD5quat4 salida );
void Quat_RotarPunto ( const MD5quat4 q, const MD5vec3 entrada, MD5vec3 salida );
void Quat_RotarPuntoNNorm ( const MD5quat4 q, const MD5vec3 entrada_nor, MD5vec3 salida_nor );
float Quat_ProductoPunto ( const MD5quat4 qa, const MD5quat4 qb );
void Quat_Interpolar ( const MD5quat4 qa, const MD5quat4 qb, float t, MD5quat4 salida );

/*
 * AbrirMD5: Abre un archivo md5mesh
 */
MD5modelo* AbrirMD5 ( const char *filename );

/*
 * FreeMD5: Libera los recursos de un modelo
 */
void FreeMD5 ( MD5modelo *modelo );

/*
 * PrepararMD5: Genera los arrays con los vertices, coordenadas de textura y normales
 */
void PrepararMD5 ( const MD5mesh *mesh,  const MD5union *esqueleto );

/*
 * MD5CalculoNormales: Calcula las normales del modelo con el baseEsqueleto
 */
void MD5CalculoNormales ( const MD5modelo *modelo, MD5mesh* mesh);

/*
 * AnimValida: Verifica si la animacion es valida para el modelo
 */
int AnimValida (const MD5modelo *modelo, const MD5anim *animacion);

/*
 * AbrirMD5Anim: Abre un archivo md5anim
 */
MD5anim* AbrirMD5Anim ( const char *filename );

/*
 * AddMD5Anim: Abre un archivo md5anim y lo aade a un modelo
 */
void AddMD5Anim ( MD5modelo *modelo, const char *filename );

/*
 * FreeAnim: Libera los recursos de una anim
 */
void FreeAnim (MD5anim *anim);

/*
 * InterpolarEsqueletos: Calcula la posicion de los vertices entre 2 posiciones interpolandolas
 */
void InterpolarEsqueletos ( const MD5union *esqueletoA, const MD5union *esqueletoB, int num_uniones, float interp, MD5union *salida );

/*
 * Animar: Calcula cual es el frame actual en funcion de dt
 */
void Animar ( const MD5anim *anim,  MD5infoanim *InfoAnim, float dt);

#endif // REVMD5_H
