////////////////////////////////////////////////
//Revolution Engine.
//by Technik
//Md5 Loading and animation code by KaWeNGoD
/////////////////////////////////////////////////

#include "REV.h"

/*
 * Operaciones basicas con quaterniones
 */
void Quat_calculoW ( MD5quat4 q )
{
	float t = 1.0f - (q[X] * q[X]) - (q[Y] * q[Y]) - (q[Z] * q[Z]);

	if ( t < 0.0f )
		q[W] = 0.0f;
	else
		q[W] = -sqrt (t);
}

void Quat_normalizar ( MD5quat4 q )
{
	// Magnitud del quaternion
	float mag = sqrt ((q[X] * q[X]) + (q[Y] * q[Y]) + (q[Z] * q[Z]) + (q[W] * q[W]));

	// Chequea tamaño erroneo para proteger contra la division entre cero
	if ( mag > 0.0f )
	{
		// Lo normalizamos
		float oneOverMag = 1.0f / mag;

		q[X] *= oneOverMag;
		q[Y] *= oneOverMag;
		q[Z] *= oneOverMag;
		q[W] *= oneOverMag;
	}
}

void Quat_multQuat ( const MD5quat4 qa, const MD5quat4 qb, MD5quat4 salida )
{
	salida[W] = (qa[W] * qb[W]) - (qa[X] * qb[X]) - (qa[Y] * qb[Y]) - (qa[Z] * qb[Z]);
	salida[X] = (qa[X] * qb[W]) + (qa[W] * qb[X]) + (qa[Y] * qb[Z]) - (qa[Z] * qb[Y]);
	salida[Y] = (qa[Y] * qb[W]) + (qa[W] * qb[Y]) + (qa[Z] * qb[X]) - (qa[X] * qb[Z]);
	salida[Z] = (qa[Z] * qb[W]) + (qa[W] * qb[Z]) + (qa[X] * qb[Y]) - (qa[Y] * qb[X]);
}

void Quat_multVec ( const MD5quat4 q, const MD5vec3 v, MD5quat4 salida )
{
	salida[W] = - (q[X] * v[X]) - (q[Y] * v[Y]) - (q[Z] * v[Z]);
	salida[X] =   (q[W] * v[X]) + (q[Y] * v[Z]) - (q[Z] * v[Y]);
	salida[Y] =   (q[W] * v[Y]) + (q[Z] * v[X]) - (q[X] * v[Z]);
	salida[Z] =   (q[W] * v[Z]) + (q[X] * v[Y]) - (q[Y] * v[X]);
}

void Quat_RotarPunto ( const MD5quat4 q, const MD5vec3 entrada, MD5vec3 salida )
{
	MD5quat4 tmp, inv, final;

	// La inversa del quaternion
	inv[X] = -q[X];
	inv[Y] = -q[Y];
	inv[Z] = -q[Z];
	inv[W] =  q[W];

	Quat_normalizar (inv);

	// La rotacion de un punto por un quaternion se hace con la formula:
	// R = Q.P.Q* ( multiplicar el quaternion por punto por la inversa del quaternion )

	Quat_multVec (q, entrada, tmp);
	Quat_multQuat (tmp, inv, final);

	salida[X] = final[X];
	salida[Y] = final[Y];
	salida[Z] = final[Z];
}

void Quat_RotarPuntoNNorm ( const MD5quat4 q, const MD5vec3 entrada_nor, MD5vec3 salida_nor )
{
	MD5quat4 tmp, inv, final;

	// La inversa del quaternion
	inv[X] = -q[X];
	inv[Y] = -q[Y];
	inv[Z] = -q[Z];
	inv[W] =  q[W];

	Quat_normalizar (inv);

	// Rotamos la normal
	Quat_multVec (q, entrada_nor, tmp);
	Quat_multQuat (tmp, inv, final);

	salida_nor[X] = final[X];
	salida_nor[Y] = final[Y];
	salida_nor[Z] = final[Z];
}

/*
 * Operaciones de los quaterniones para la animacion de esqueletos
 */
float Quat_ProductoPunto (const MD5quat4 qa, const MD5quat4 qb)
{
	return ((qa[X] * qb[X]) + (qa[Y] * qb[Y]) + (qa[Z] * qb[Z]) + (qa[W] * qb[W]));
}

void Quat_Interpolar (const MD5quat4 qa, const MD5quat4 qb, float t, MD5quat4 salida)
{
	// Verificamos si los parametros estan fuera de rango

	if ( t <= 0.0 )
	{
		memcpy (salida, qa, sizeof(MD5quat4));
		return;
	}

	if ( t >= 1.0 )
	{
		memcpy (salida, qb, sizeof (MD5quat4));
		return;
	}

	// Calculamos el "coseno del angulo entre quaterniones" usando el producto de punto
	float cosOmega = Quat_ProductoPunto (qa, qb);

	// Si el punto es negativo usamos -q1.  q y -q representan la misma rotacion pero puede
	// producir un resultado diferente. Elegimos q o -q para rotar usando el angulo
	float q1w = qb[W];
	float q1x = qb[X];
	float q1y = qb[Y];
	float q1z = qb[Z];

	if ( cosOmega < 0.0f )
	{
		q1w = -q1w;
		q1x = -q1x;
		q1y = -q1y;
		q1z = -q1z;
		cosOmega = -cosOmega;
	}

	// Debemos tener 2 unidades, entonces el punto no puede ser <= 1.0
//	if ( cosOmega < 1.1f );
//		exit(0);

	// Calculamos la interpolacion, buscando quaterniones que son iguales
	float k0, k1;

	if ( cosOmega > 0.9999f )
	{
		// Muy cerca, usamos interpolacion linear
		k0 = 1.0f - t;
		k1 = t;
	}
	else
	{
		// Calculamos el seno del angulo
		float sinOmega = sqrt (1.0f - ( cosOmega * cosOmega ) );

		// Calculamos el angulo del seno y coseno
		float omega = atan2 ( sinOmega, cosOmega );

		// Calculamos el denominador inverso
		float oneOverSinOmega = 1.0f / sinOmega;

		// Calculamos los parametros de interpolacion
		k0 = sin ((1.0f - t) * omega) * oneOverSinOmega;
		k1 = sin (t * omega) * oneOverSinOmega;
	}

	// Interpolamos y devolvemos un nuevo quaternion
	salida[W] = (k0 * qa[3]) + (k1 * q1w);
	salida[X] = (k0 * qa[0]) + (k1 * q1x);
	salida[Y] = (k0 * qa[1]) + (k1 * q1y);
	salida[Z] = (k0 * qa[2]) + (k1 * q1z);
}

/*
 * Carga un modelo MD5 desde un archivo
 */
MD5modelo* AbrirMD5 ( const char *filename )
{
	FILE *fp;
	MD5modelo *modelo = NULL;
	char buff[512];
	int version;
	int mesh_actual = 0;
	int i;

	fp = fopen (filename, "rb");
	if ( !fp )
	{
		fprintf (stderr, "Error: no se puede abrir \"%s\"!\n", filename);
		return NULL;
	}

	// Reservamos memoria para el modelo
	modelo = (MD5modelo*)memalign(32,sizeof(MD5modelo));

	while ( !feof ( fp ) )
	{
		// Leemos una linea entera
		fgets ( buff, sizeof ( buff ), fp );

		if ( sscanf ( buff, " MD5Version %d", &version ) == 1 )
		{
			if ( version != MD5_VERSION )
			{
				// Version incorrecta
				fprintf (stderr, "Error: version del modelo incorrecta\n");
				free( modelo );
				fclose (fp);
				return NULL;
			}
		}
		else if ( sscanf ( buff, " numJoints %d", &modelo->num_uniones ) == 1 )
		{
			if ( modelo->num_uniones > 0 )
			{
				// Reservamos memoria para el esqueleto base de uniones
				modelo->baseEsqueleto = (MD5union*)memalign(32,sizeof(MD5union) * modelo->num_uniones );
			}
		}
		else if ( sscanf ( buff, " numMeshes %d", &modelo->num_meshes) == 1 )
		{
			if (modelo->num_meshes > 0)
			{
				// Reservamos memoria para los meshes
				modelo->meshes = (MD5mesh*)memalign(32,sizeof(MD5mesh) * modelo->num_meshes );
			}
		}
		else if ( strncmp ( buff, "joints {", 8) == 0 )
		{
			// Lee cada union
			for ( i = 0; i < modelo->num_uniones; ++i )
			{
				MD5union *uniones = &modelo->baseEsqueleto[i];

				// Leemos una linea entera
				fgets ( buff, sizeof(buff), fp );

				if ( sscanf ( buff, "%s %d ( %f %f %f ) ( %f %f %f )",
					uniones->nombre, &uniones->pariente, &uniones->pos[0],
					&uniones->pos[1], &uniones->pos[2], &uniones->orient[0],
					&uniones->orient[1], &uniones->orient[2]) == 8 )
				{
					// Calcula el componente w
					Quat_calculoW (uniones->orient);
				}
			}
		}
		else if ( strncmp ( buff, "mesh {", 6) == 0 )
		{
			MD5mesh *mesh = &modelo->meshes[mesh_actual];
			int vert_index = 0;
			int tri_index = 0;
			int peso_index = 0;
			float fdata[4];
			int idata[3];

			while ( ( buff[0] != '}' ) && !feof ( fp ) )
			{
				// Leemos una linea entera
				fgets ( buff, sizeof(buff), fp );

				if ( strstr ( buff, "shader " ) )
				{
					int quote = 0, j = 0;

					// Copia el nombre del shader sin quotes
					for ( i = 0; i < sizeof(buff) && ( quote < 2 ); ++i )
					{
						if ( buff[i] == '\"' )
							quote++;

						if ( ( quote == 1 ) && ( buff[i] != '\"' ) )
						{
							mesh->shader[j] = buff[i];
							j++;
						}
					}
				}
				else if ( sscanf ( buff, " numverts %d", &mesh->num_vertices ) == 1 )
				{
					if ( mesh->num_vertices > 0 )
					{
						// Reserva memoria para los vertices
						mesh->vertices = (MD5vertice*)memalign(32,sizeof(MD5vertice) * mesh->num_vertices );
					}
				}
				else if ( sscanf ( buff, " numtris %d", &mesh->num_triangulos ) == 1 )
				{
					if (mesh->num_triangulos > 0)
					{
						// Reservamos memomria para los triangulos
						mesh->triangulos = (MD5triangulo*)memalign(32,sizeof(MD5triangulo) * mesh->num_triangulos );
					}
				}
				else if ( sscanf ( buff, " numweights %d", &mesh->num_pesos ) == 1 )
				{
					if ( mesh->num_pesos > 0 )
					{
						// Reservamos memomria para los pesos
						mesh->pesos = (MD5peso*)memalign(32,sizeof(MD5peso) * mesh->num_pesos );
					}
				}
				else if ( sscanf ( buff, " vert %d ( %f %f ) %d %d", &vert_index, &fdata[0], &fdata[1], &idata[0], &idata[1] ) == 5 )
				{
					// Copiamos los datos de los vertices
					mesh->vertices[vert_index].st[0]  = fdata[0];
					mesh->vertices[vert_index].st[1]  = fdata[1];
					mesh->vertices[vert_index].inicio = idata[0];
					mesh->vertices[vert_index].total  = idata[1];
				}
				else if ( sscanf ( buff, " tri %d %d %d %d", &tri_index, &idata[0], &idata[1], &idata[2] ) == 4 )
				{
					// Copiamos los datos de los triangulos
					mesh->triangulos[tri_index ].indice[0] = idata[0];
					mesh->triangulos[tri_index ].indice[1] = idata[1];
					mesh->triangulos[tri_index ].indice[2] = idata[2];
				}
				else if ( sscanf ( buff, " weight %d %d %f ( %f %f %f )", &peso_index, &idata[0], &fdata[3], &fdata[0], &fdata[1], &fdata[2] ) == 6 )
				{
					// Copiamos los datos de los vertices
					mesh->pesos[peso_index].uniones  = idata[0];
					mesh->pesos[peso_index].bias     = fdata[3];
					mesh->pesos[peso_index].pos[0]   = fdata[0];
					mesh->pesos[peso_index].pos[1]   = fdata[1];
					mesh->pesos[peso_index].pos[2]   = fdata[2];
				}

			}

			mesh_actual++;
		}
	}

	modelo->animacion = NULL;
	modelo->num_animaciones = 0;

	modelo->Infoanim = (MD5infoanim*)memalign( 32, sizeof(MD5infoanim));

	modelo->Infoanim->curr_frame = 0;
	modelo->Infoanim->next_frame = 1;
	modelo->Infoanim->last_time = 0;
	modelo->Infoanim->max_time = 1.0;

	// Reservamos memoria para los arrays e indices del modelo y generamos las normales iniciales
	for( i=0; i < modelo->num_meshes; i++)
	{
		MD5mesh *mesh = &modelo->meshes[i];

		mesh->Verts 	= (f32*)memalign(32,sizeof(f32) * 3 * mesh->num_vertices);
		mesh->Texcoords	= (f32*)memalign(32,sizeof(f32) * 2 * mesh->num_vertices);
		mesh->Normales	= (f32*)memalign(32,sizeof(f32) * 3 * mesh->num_vertices);

		mesh->VIndices	= (u16*)memalign(32,sizeof(u16) * 3 * mesh->num_triangulos);

		// Calculamos las normales iniciales
		MD5CalculoNormales ( modelo, mesh);
	}

	fclose ( fp );

	return modelo;
}

/*
 * Libera los recursos reservados para el modelo
 */
void FreeMD5 ( MD5modelo *modelo )
{
	int i;

	if ( modelo->baseEsqueleto )
	{
		free ( modelo->baseEsqueleto );
		modelo->baseEsqueleto = NULL;
	}

	if ( modelo->meshes )
	{
		// Liberamos los datos del mesh
		for ( i = 0; i < modelo->num_meshes; ++i )
		{
			if ( modelo->meshes[i].vertices )
			{
				free ( modelo->meshes[i].vertices );
				modelo->meshes[i].vertices = NULL;
			}

			if ( modelo->meshes[i].triangulos )
			{
				free ( modelo->meshes[i].triangulos );
				modelo->meshes[i].triangulos = NULL;
			}

			if ( modelo->meshes[i].pesos )
			{
				free ( modelo->meshes[i].pesos );
				modelo->meshes[i].pesos = NULL;
			}

			if ( modelo->meshes[i].norm_inic )
			{
				free ( modelo->meshes[i].norm_inic );
				modelo->meshes[i].norm_inic = NULL;
			}

			if ( modelo->meshes[i].Verts )
			{
				free ( modelo->meshes[i].Verts );
				modelo->meshes[i].Verts = NULL;
			}

			if ( modelo->meshes[i].Texcoords )
			{
				free ( modelo->meshes[i].Texcoords );
				modelo->meshes[i].Texcoords = NULL;
			}

			if ( modelo->meshes[i].Normales )
			{
				free ( modelo->meshes[i].Normales );
				modelo->meshes[i].Normales = NULL;
			}

			if ( modelo->meshes[i].VIndices )
			{
				free ( modelo->meshes[i].VIndices );
				modelo->meshes[i].VIndices = NULL;
			}
		}

		free ( modelo->meshes );
		modelo->meshes = NULL;
	}

	if ( modelo->animacion )
	{
		// Liberamos los datos de las animaciones
		for ( i = 0; i < modelo->num_animaciones; ++i )
			FreeAnim(&modelo->animacion[i]);

		free ( modelo->animacion );
		modelo->animacion = NULL;
	}

	if ( modelo->Infoanim )
	{
		free( modelo->Infoanim );
		modelo->Infoanim = NULL;
	}

	free ( modelo );
	modelo = NULL;
}

/*
 * Prepara un mesh para dibujarlo. Calcula las posiciones finales de los vertices
 * en relacion al esqueleto.
 */
void PrepararMD5 ( const MD5mesh *mesh, const MD5union *esqueleto )
{
	int i, j, k;

	// Configura los indices de vertices
	for ( k = 0, i = 0; i < mesh->num_triangulos; ++i )
	{
		for ( j = 0; j < 3; ++j, ++k )
			mesh->VIndices[k] = mesh->triangulos[i].indice[j];
	}

	// Configuramos los vertices
	for ( i = 0; i < mesh->num_vertices; ++i )
	{
		MD5vec3 Verticefinal = { 0.0f, 0.0f, 0.0f };
		MD5vec3 new_norm = { 0.0f, 0.0f, 0.0f };
		MD5vec3 tmp_norm;

		tmp_norm[0] = mesh->norm_inic[i][0];
		tmp_norm[1] = mesh->norm_inic[i][1];
		tmp_norm[2] = mesh->norm_inic[i][2];

		mesh->Texcoords[(i*2)+0] = mesh->vertices[i].st[0];
		mesh->Texcoords[(i*2)+1] = 1.0-mesh->vertices[i].st[1];

		// Calcula el vertice final para dibujar con pesos
		for ( j = 0; j < mesh->vertices[i].total; ++j )
		{
			const MD5peso *peso = &mesh->pesos[mesh->vertices[i].inicio + j];
			const MD5union *uniones = &esqueleto[peso->uniones];

			// Calcula el vertice transformado para este peso
			MD5vec3 wv;
			MD5vec3 wn;

			Quat_RotarPunto (uniones->orient, peso->pos, wv);
			Quat_RotarPuntoNNorm (uniones->orient, tmp_norm, wn);

			// La suma de todos los peso->bias debe ser 1.0
			Verticefinal[0] += (uniones->pos[0] + wv[0]) * peso->bias;
			Verticefinal[1] += (uniones->pos[1] + wv[1]) * peso->bias;
			Verticefinal[2] += (uniones->pos[2] + wv[2]) * peso->bias;

			new_norm[0] += wn[0] * peso->bias;
			new_norm[1] += wn[1] * peso->bias;
			new_norm[2] += wn[2] * peso->bias;
		}

		mesh->Verts[(i*3)+0] = Verticefinal[0];
		mesh->Verts[(i*3)+1] = Verticefinal[1];
		mesh->Verts[(i*3)+2] = Verticefinal[2];

		mesh->Normales[(i*3)+0] = new_norm[0];
		mesh->Normales[(i*3)+1] = new_norm[1];
		mesh->Normales[(i*3)+2] = new_norm[2];
	}

}

/*
 * AnimValida: Verifica si la animacion es valida para el modelo
 */
int AnimValida (const MD5modelo *modelo, const MD5anim *animacion)
{
	int i;

	// md5mesh y md5anim deben tener el mismo numero de uniones
	if ( modelo->num_uniones != animacion->num_uniones )
		return 0;

	// Verificamos con el frame[0]
	for ( i = 0; i < modelo->num_uniones; ++i )
	{
		// Las uniones deben tener el mismo indice de pariente
		if (modelo->baseEsqueleto[i].pariente != animacion->EsqueletoFrames[0][i].pariente)
		return 0;

		// Las uniones deben tener el mismo nombre
		if (strcmp ( modelo->baseEsqueleto[i].nombre, animacion->EsqueletoFrames[0][i].nombre ) != 0)
			return 0;
	}

	return 1;
}

/*
 * CrearEsqueletoFrame: Crea un esqueleto para el frame dado
 */
static void CrearEsqueletoFrame ( const MD5infounion *Infounion, const MD5unionbase *baseFrame, const float *animFrameData, MD5union *esqueletoFrame, int num_uniones )
{
	int i;

	for ( i = 0; i < num_uniones; ++i )
	{
		const MD5unionbase *UnionBase = &baseFrame[i];
		MD5vec3 PosAnimada;
		MD5quat4 OrientAnimada;
		int j = 0;

		memcpy (PosAnimada, UnionBase->pos, sizeof (MD5vec3));
		memcpy (OrientAnimada, UnionBase->orient, sizeof (MD5quat4));

		if (Infounion[i].flags & 1) // Tx
		{
			PosAnimada[0] = animFrameData[Infounion[i].IndiceInicio + j];
			++j;
		}
		if (Infounion[i].flags & 2) // Ty
		{
			PosAnimada[1] = animFrameData[Infounion[i].IndiceInicio + j];
			++j;
		}
		if (Infounion[i].flags & 4) // Tz
		{
			PosAnimada[2] = animFrameData[Infounion[i].IndiceInicio + j];
			++j;
		}
		if (Infounion[i].flags & 8) // Qx
		{
			OrientAnimada[0] = animFrameData[Infounion[i].IndiceInicio + j];
			++j;
		}
		if (Infounion[i].flags & 16) // Qy
		{
			OrientAnimada[1] = animFrameData[Infounion[i].IndiceInicio + j];
			++j;
		}
		if (Infounion[i].flags & 32) // Qz
		{
			OrientAnimada[2] = animFrameData[Infounion[i].IndiceInicio + j];
			++j;
		}

		// Calculamos el valor w de OrientAnimada
		Quat_calculoW ( OrientAnimada );

		MD5union *estaUnion = &esqueletoFrame[i];

		int pariente = Infounion[i].pariente;
		estaUnion->pariente = pariente;
		strcpy (estaUnion->nombre, Infounion[i].nombre);

		// Tiene pariente
		if (estaUnion->pariente < 0)
		{
			memcpy (estaUnion->pos, PosAnimada, sizeof(MD5vec3));
			memcpy (estaUnion->orient, OrientAnimada, sizeof(MD5quat4));
		}
		else
		{
			MD5union *Unionpariente = &esqueletoFrame[pariente];
			MD5vec3 rpos; // Posicion rotada

			// Añadimos posiciones
			Quat_RotarPunto (Unionpariente->orient, PosAnimada, rpos);
			estaUnion->pos[0] = rpos[0] + Unionpariente->pos[0];
			estaUnion->pos[1] = rpos[1] + Unionpariente->pos[1];
			estaUnion->pos[2] = rpos[2] + Unionpariente->pos[2];

			// Concatenamos rotaciones
			Quat_multQuat (Unionpariente->orient, OrientAnimada, estaUnion->orient);
			Quat_normalizar (estaUnion->orient);
		}
	}
}

/*
 * AbrirMD5Anim: Carga una animacion MD5
 */
MD5anim* AbrirMD5Anim ( const char *filename )
{
	FILE *fp = NULL;
	char buff[512];
	MD5anim *animacion = NULL;
	MD5infounion *Infounion = NULL;
	MD5unionbase *baseFrame = NULL;
	float *animFrameData = NULL;
	int version;
	int numCompsAnimados;
	int indiceframe;
	int i;

	fp = fopen (filename, "rb");
	if (!fp)
	{
		fprintf (stderr, "error: ni se puede abrir \"%s\"!\n", filename);
		return 0;
	}

	// Reservamos memoria para la animacion
	animacion = (MD5anim*)memalign(32,sizeof(MD5anim));

	while (!feof (fp))
	{

		// Cojemos una linea
		fgets (buff, sizeof (buff), fp);

		if (sscanf (buff, " MD5Version %d", &version) == 1)
		{
			if (version != 10)
			{
				// Version incorrecta
				fprintf (stderr, "Error: version incorrecta\n");
				free( animacion );
				fclose (fp);
				return NULL;
			}
		}
		else if (sscanf (buff, " numFrames %d", &animacion->num_frames) == 1)
		{
			// Reservamos memoria para las cajas y frames del esqueleto
			if (animacion->num_frames > 0)
			{
				animacion->EsqueletoFrames = (MD5union **)memalign(32, sizeof(MD5union*) * animacion->num_frames);
				animacion->cajas = (MD5caja *)memalign(32, sizeof(MD5caja) * animacion->num_frames);
			}
		}
		else if (sscanf (buff, " numJoints %d", &animacion->num_uniones) == 1)
		{
			if (animacion->num_uniones > 0)
			{
				for (i = 0; i < animacion->num_frames; ++i)
				{
					// Reservamos memoria para las uniones de cada frame
					animacion->EsqueletoFrames[i] = (MD5union *)memalign(32, sizeof(MD5union) * animacion->num_uniones);
				}

				// Reservamos memoria temporal para contruir los frames del esqueleto
				Infounion = (MD5infounion *)memalign(32, sizeof(MD5infounion) * animacion->num_uniones);

				baseFrame = (MD5unionbase *)memalign(32, sizeof(MD5unionbase) * animacion->num_uniones);
			}
		}
		else if (sscanf (buff, " frameRate %d", &animacion->frameRate) == 1)
		{
			printf ("AbrirMD5Anim: El frame rate de la animacion es %d\n", animacion->frameRate);
		}
		else if (sscanf (buff, " numAnimatedComponents %d", &numCompsAnimados) == 1)
		{
			if (numCompsAnimados > 0)
			{
				// Reservamos memoria para los datos de animacion del frame
				animFrameData = (float *)memalign(32, sizeof(float) * numCompsAnimados);
			}
		}
		else if (strncmp (buff, "hierarchy {", 11) == 0)
		{
			for (i = 0; i < animacion->num_uniones; ++i)
			{
				// Leemos una linea
				fgets (buff, sizeof (buff), fp);

				// Leemos informacion de union
				sscanf (buff, " %s %d %d %d", Infounion[i].nombre, &Infounion[i].pariente, &Infounion[i].flags, &Infounion[i].IndiceInicio);
			}
		}
		else if (strncmp (buff, "bounds {", 8) == 0)
		{
			for (i = 0; i < animacion->num_frames; ++i)
			{
				// Leemos una linea
				fgets (buff, sizeof (buff), fp);

				// Leemos las cajas
				sscanf (buff, " ( %f %f %f ) ( %f %f %f )",
						&animacion->cajas[i].min[0], &animacion->cajas[i].min[1],
						&animacion->cajas[i].min[2], &animacion->cajas[i].max[0],
						&animacion->cajas[i].max[1], &animacion->cajas[i].max[2]);
			}
		}
		else if (strncmp (buff, "baseframe {", 10) == 0)
		{
			for (i = 0; i < animacion->num_uniones; ++i)
			{
				// Leemos una linea
				fgets (buff, sizeof (buff), fp);

				// Leemos el frame base de la union
				if (sscanf (buff, " ( %f %f %f ) ( %f %f %f )",
						&baseFrame[i].pos[0], &baseFrame[i].pos[1],
						&baseFrame[i].pos[2], &baseFrame[i].orient[0],
						&baseFrame[i].orient[1], &baseFrame[i].orient[2]) == 6)
				{
					// Calculamos el componente w
					Quat_calculoW (baseFrame[i].orient);
				}
			}
		}
		else if (sscanf (buff, " frame %d", &indiceframe) == 1)
		{
			// Leemos los datos del frame
			for (i = 0; i < numCompsAnimados; ++i)
				fscanf (fp, "%f", &animFrameData[i]);

			// Construimos el esqueleto del frame con los datos recopilados
			CrearEsqueletoFrame (Infounion, baseFrame, animFrameData, animacion->EsqueletoFrames[indiceframe], animacion->num_uniones);
		}
	}

	fclose (fp);
	
	// Liberamos memoria temporal
	if (animFrameData)
		free (animFrameData);

	if (baseFrame)
		free (baseFrame);

	if (Infounion)
		free (Infounion);

	return animacion;
}

/*
 * AddMD5Anim: abre una animacion y la añade una animacion aun modelo MD5
 */
void AddMD5Anim ( MD5modelo *modelo, const char *filename )
{
	FILE *fp = NULL;
	char buff[512];
	MD5anim *animacion = NULL;
	MD5infounion *Infounion = NULL;
	MD5unionbase *baseFrame = NULL;
	float *animFrameData = NULL;
	int version;
	int numCompsAnimados;
	int indiceframe;
	int i;

	fp = fopen (filename, "rb");
	if (!fp)
	{
		fprintf (stderr, "error: ni se puede abrir \"%s\"!\n", filename);
		return;
	}

	// Reservamos memoria para la animacion
	if ( modelo->animacion == NULL )
		modelo->animacion = (MD5anim*)memalign( 32,sizeof(MD5anim) );
	else
		realloc(modelo->animacion, sizeof(MD5anim) * ( modelo->num_animaciones + 1 ));

	// Cojemos la animacion nueva
	animacion = &modelo->animacion[modelo->num_animaciones];

	while (!feof (fp))
	{

		// Cojemos una linea
		fgets (buff, sizeof (buff), fp);

		if (sscanf (buff, " MD5Version %d", &version) == 1)
		{
			if (version != 10)
			{
				// Version incorrecta
				fprintf (stderr, "Error: version incorrecta\n");
				free( animacion );
				fclose (fp);
				return;
			}
		}
		else if (sscanf (buff, " numFrames %d", &animacion->num_frames) == 1)
		{
			// Reservamos memoria para las cajas y frames del esqueleto
			if (animacion->num_frames > 0)
			{
				animacion->EsqueletoFrames = (MD5union **)memalign(32, sizeof(MD5union*) * animacion->num_frames);
				animacion->cajas = (MD5caja *)memalign(32, sizeof(MD5caja) * animacion->num_frames);
			}
		}
		else if (sscanf (buff, " numJoints %d", &animacion->num_uniones) == 1)
		{
			if (animacion->num_uniones > 0)
			{
				for (i = 0; i < animacion->num_frames; ++i)
				{
					// Reservamos memoria para las uniones de cada frame
					animacion->EsqueletoFrames[i] = (MD5union *)memalign(32, sizeof(MD5union) * animacion->num_uniones);
				}

				// Reservamos memoria temporal para contruir los frames del esqueleto
				Infounion = (MD5infounion *)memalign(32, sizeof(MD5infounion) * animacion->num_uniones);

				baseFrame = (MD5unionbase *)memalign(32, sizeof(MD5unionbase) * animacion->num_uniones);
			}
		}
		else if (sscanf (buff, " frameRate %d", &animacion->frameRate) == 1)
		{
//			printf ("AbrirMD5Anim: El frame rate de la animacion es %d\n", animacion->frameRate);
		}
		else if (sscanf (buff, " numAnimatedComponents %d", &numCompsAnimados) == 1)
		{
			if (numCompsAnimados > 0)
			{
				// Reservamos memoria para los datos de animacion del frame
				animFrameData = (float *)memalign(32, sizeof(float) * numCompsAnimados);
			}
		}
		else if (strncmp (buff, "hierarchy {", 11) == 0)
		{
			for (i = 0; i < animacion->num_uniones; ++i)
			{
				// Leemos una linea
				fgets (buff, sizeof (buff), fp);

				// Leemos informacion de union
				sscanf (buff, " %s %d %d %d", Infounion[i].nombre, &Infounion[i].pariente, &Infounion[i].flags, &Infounion[i].IndiceInicio);
			}
		}
		else if (strncmp (buff, "bounds {", 8) == 0)
		{
			for (i = 0; i < animacion->num_frames; ++i)
			{
				// Leemos una linea
				fgets (buff, sizeof (buff), fp);

				// Leemos las cajas
				sscanf (buff, " ( %f %f %f ) ( %f %f %f )",
						&animacion->cajas[i].min[0], &animacion->cajas[i].min[1],
						&animacion->cajas[i].min[2], &animacion->cajas[i].max[0],
						&animacion->cajas[i].max[1], &animacion->cajas[i].max[2]);
			}
		}
		else if (strncmp (buff, "baseframe {", 10) == 0)
		{
			for (i = 0; i < animacion->num_uniones; ++i)
			{
				// Leemos una linea
				fgets (buff, sizeof (buff), fp);

				// Leemos el frame base de la union
				if (sscanf (buff, " ( %f %f %f ) ( %f %f %f )",
						&baseFrame[i].pos[0], &baseFrame[i].pos[1],
						&baseFrame[i].pos[2], &baseFrame[i].orient[0],
						&baseFrame[i].orient[1], &baseFrame[i].orient[2]) == 6)
				{
					// Calculamos el componente w
					Quat_calculoW (baseFrame[i].orient);
				}
			}
		}
		else if (sscanf (buff, " frame %d", &indiceframe) == 1)
		{
			// Leemos los datos del frame
			for (i = 0; i < numCompsAnimados; ++i)
				fscanf (fp, "%f", &animFrameData[i]);

			// Construimos el esqueleto del frame con los datos recopilados
			CrearEsqueletoFrame (Infounion, baseFrame, animFrameData, animacion->EsqueletoFrames[indiceframe], animacion->num_uniones);
		}
	}

	fclose (fp);
	
	// Liberamos memoria temporal
	if (animFrameData)
		free (animFrameData);

	if (baseFrame)
		free (baseFrame);

	if (Infounion)
		free (Infounion);

	modelo->num_animaciones++;

	modelo->Infoanim->curr_frame = 0;
	modelo->Infoanim->next_frame = 1;
	modelo->Infoanim->last_time = 0;
	modelo->Infoanim->max_time = (float)(1.0 / (float)animacion->frameRate);

	return;
}

/*
 * Libera los recursos reservados para la animacion.
 */
void FreeAnim (MD5anim *anim)
{
	int i;

	if (anim->EsqueletoFrames)
	{
		for ( i = 0; i < anim->num_frames; ++i )
		{
			if ( anim->EsqueletoFrames[i] )
			{
				free (anim->EsqueletoFrames[i]);
				anim->EsqueletoFrames[i] = NULL;
			}
		}

		free (anim->EsqueletoFrames);
		anim->EsqueletoFrames = NULL;
	}

	if ( anim->cajas )
	{
		free (anim->cajas);
		anim->cajas = NULL;
	}

	free( anim );
}

/*
 * Interpolacion suave entre 2 esqueletos
 */
void InterpolarEsqueletos (const MD5union *esqueletoA, const MD5union *esqueletoB, int num_uniones, float interp, MD5union *salida )
{
	int i;

	for (i = 0; i < num_uniones; ++i)
	{

		// copia el indice pariente
		salida[i].pariente = esqueletoA[i].pariente;

		// Interpolacion lineal de la posicion
		salida[i].pos[0] = esqueletoA[i].pos[0] + interp * (esqueletoB[i].pos[0] - esqueletoA[i].pos[0]);
		salida[i].pos[1] = esqueletoA[i].pos[1] + interp * (esqueletoB[i].pos[1] - esqueletoA[i].pos[1]);
		salida[i].pos[2] = esqueletoA[i].pos[2] + interp * (esqueletoB[i].pos[2] - esqueletoA[i].pos[2]);

		// Interpolacion para la orientacion
		Quat_Interpolar (esqueletoA[i].orient, esqueletoB[i].orient, interp, salida[i].orient);
	}
}

/*
 * Realiza calculos de la animacion. Calcula el frame actual y siguiente en funcion del tiempo delta
 */
void Animar (const MD5anim *anim, MD5infoanim *Infoanim, float dt)
{
	int maxFrames = anim->num_frames - 1;

	Infoanim->last_time += dt;

	// Nos movemos al siguiente frame
	if (Infoanim->last_time >= Infoanim->max_time)
	{
		Infoanim->curr_frame++;
		Infoanim->next_frame++;
		Infoanim->last_time = 0.0;

		if (Infoanim->curr_frame > maxFrames)
			Infoanim->curr_frame = 0;

		if (Infoanim->next_frame > maxFrames)
			Infoanim->next_frame = 0;
	}
}

void ProductoCruzado(MD5vec3 *producto, MD5vec3 a, MD5vec3 b)
{
	(*producto)[0] = (a[1]*b[2]) - (a[2]*b[1]);
	(*producto)[1] = (a[2]*b[0]) - (a[0]*b[2]);
	(*producto)[2] = (a[0]*b[1]) - (a[1]*b[0]);
}

float ProductoPunto(MD5vec3 a, MD5vec3 b)
{
	return (a[0]*b[0])+(a[1]*b[1])+(a[2]*b[2]);
}

void Normalizar (MD5vec3* vector)
{
	float mag = sqrt( ProductoPunto((*vector),(*vector)) );

	(*vector)[0] /= mag;
	(*vector)[1] /= mag;
	(*vector)[2] /= mag;
}

/*
 * Obtenemos las normales del esqueleto base de un MD5
 */
void MD5CalculoNormales ( const MD5modelo *modelo, MD5mesh* mesh)
{
	int i = 0, j = 0;

	//work out smooth shading pervertex normals
	mesh->norm_inic = (MD5vec3*)memalign(32, sizeof(MD5vec3) * mesh->num_vertices );
	MD5vec3 *Vertices = (MD5vec3*)memalign(32, sizeof(MD5vec3) * mesh->num_vertices );

	// Calculamos las posiciones finales de los vértices
	for ( i = 0; i < mesh->num_vertices; ++i )
	{
		MD5vec3 Verticefinal = { 0.0f, 0.0f, 0.0f };

		// Calcula el vertice final para dibujar con pesos
		for ( j = 0; j < mesh->vertices[i].total; ++j )
		{
			const MD5peso *peso = &mesh->pesos[mesh->vertices[i].inicio + j];
			const MD5union *uniones = modelo->baseEsqueleto;

			// Calcula el vertice transformado para este peso
			MD5vec3 wv;
			Quat_RotarPunto (uniones->orient, peso->pos, wv);

			// La suma de todos los peso->bias debe ser 1.0
			Verticefinal[0] += (uniones->pos[0] + wv[0]) * peso->bias;
			Verticefinal[1] += (uniones->pos[1] + wv[1]) * peso->bias;
			Verticefinal[2] += (uniones->pos[2] + wv[2]) * peso->bias;
		}

		Vertices[i][0] = Verticefinal[0];
		Vertices[i][1] = Verticefinal[1];
		Vertices[i][2] = Verticefinal[2];
	}

	// Reservamos memoria para las normales de las caras
	MD5vec3* Fnormales = (MD5vec3*)memalign(32, sizeof(MD5vec3) * mesh->num_triangulos );

	for ( i=0; i < mesh->num_triangulos; i++ )
	{
		MD5vec3 resultado;
		MD5vec3 ab;
		MD5vec3 ac;
		MD5vec3 a,b,c;

		// Vertice 1
		a[0] = Vertices[mesh->triangulos[j].indice[0]][0];
		a[1] = Vertices[mesh->triangulos[j].indice[0]][1];
		a[2] = Vertices[mesh->triangulos[j].indice[0]][2];

		// Verice 2
		b[0] = Vertices[mesh->triangulos[j].indice[1]][0];
		b[1] = Vertices[mesh->triangulos[j].indice[1]][1];
		b[2] = Vertices[mesh->triangulos[j].indice[1]][2];

		// Vertice 3 
		c[0] = Vertices[mesh->triangulos[j].indice[2]][0];
		c[1] = Vertices[mesh->triangulos[j].indice[2]][1];
		c[2] = Vertices[mesh->triangulos[j].indice[2]][2];

		// Calcula el vector normal al poligono
		ab[0] = b[0] - a[0];
		ab[1] = b[1] - a[1];
		ab[2] = b[2] - a[2];
		ac[0] = c[0] - a[0];
		ac[1] = c[1] - a[1];
		ac[2] = c[2] - a[2];

		ProductoCruzado(&resultado,ab,ac);

		Normalizar( &resultado );

		Fnormales[i][0] = resultado[0];
		Fnormales[i][1] = resultado[1];
		Fnormales[i][2] = resultado[2];
	}

	for ( i=0; i < mesh->num_vertices; i++)
	{
		MD5vec3 comp_normal = {0.0f,0.0f,0.0f};
		float num_comp = 0;

		for ( j=0; j < mesh->num_triangulos; j++)
		{
			// Buscamos si un triangulo tiene este vertice y añadimos su normal
			if(i == mesh->triangulos[j].indice[0] || i == mesh->triangulos[j].indice[1] || i == mesh->triangulos[j].indice[2])
			{
				comp_normal[0] += Fnormales[j][0];
				comp_normal[1] += Fnormales[j][1];
				comp_normal[2] += Fnormales[j][2];
				num_comp++;
			}
		}

		comp_normal[0] /= num_comp;
		comp_normal[1] /= num_comp;
		comp_normal[2] /= num_comp;

		// Normalizamos
		Normalizar( &comp_normal );

		mesh->norm_inic[i][0] = comp_normal[0];
		mesh->norm_inic[i][1] = comp_normal[1];
		mesh->norm_inic[i][2] = comp_normal[2];

		comp_normal[0] = comp_normal[1] = comp_normal[2] = num_comp =0.0;
	}

	// Liberamos la memoria temporal
	free(Vertices);
	free(Fnormales);

}

